#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "alloc.h"
#include "p1_common.h"
#include "p1_parser.h"
#include "p1_scanner.h"

char *off_exp1();
char *off_exp2();

/* ===========================================================================

   function name: all_
   input	: ---
   output	: ---
   return value : return_value	- SUCCESS, if no errors
			 	- FAIL, otherwise

   global vars  : symbol_table	- array containing symbol table
		  get_count	- counter for making 'get' statement labels
		  token_type	- type of current token
		  token_buf	- character string for current token
		  debug		- flag for debugging statements

   description	: This routine is part of a recursive descent parser and
		  handles the following productions from the grammar:

			GET_ <-- index in LOG_EXP   BODY1 elsenany BODY2 endget   index ;
			NEXT_<-- index in MORE_NEXT BODY1 elsenany BODY2 endnext  index ;
		   FOR_ <-- index in LOG_EXP   BODY1 elsenany BODY2 endfor   index ;
			WHILE <--index in LOG_EXP   BODY1 elsenany BODY2 endwhile index ;

		  The index variable must be an identifier that has been
		  declared as an 'index parallel' identifier. The expression
		  must evaluate to a logical parallel result. The body
		  of the 'all' statement can be empty. The keywords 'in'
		  and 'endall' are required. The index variable must be
		  repeated at the end of the 'all' statement. The statements
		  in the body are executed only once. The expression is
		  evaluated only once and the result is stored in the index
		  variable. The index variable may not be modified, but
		  may be used to index a parallel variable.

=========================================================================== */

int all_(stype)
int stype; /* T_GET T_NEXT */
{
   int return_value;		/* return value of function */
   int invar_index;		/* symbol table index of index variable */
   int invar_offset;
   int expr_index;		/* symbol table index of expression result */
   int expr_offset;
   int arg1_index;
	int arg1_offset;
   char *start_label;	/* label to print at start of 'get' */
   char *else_label;    /* label to print at elsenany     */
   char *end_label;		/* label to print at end of 'get' */
   char *repeat_label;  /* label for loop to top */
   if (debug == 2)
      printf ("In all_(), stype=%d..............................\n",
				  stype);

   return_value = SUCCESS;

   start_label  = make_label ("GE$S", ++get_count);
   else_label   = make_label ("GE$L",   get_count);
	repeat_label = make_label ("GE$R",   get_count);
   end_label    = make_label ("GE$E",   get_count);

   print_opcode ("setscope", T_SETSCOPE, T_NONE, T_NONE);
   print_args (start_label, NULL, NULL, NULL, NULL);

   if(stype == T_WHILE)
	  {print_opcode("begwhile", T_BEGWHILE, T_NONE, T_NONE);
		print_args(repeat_label, NULL, NULL, NULL, NULL);}

   get_index(&invar_index,&invar_offset);
   if (invar_index == FAIL) return_value = FAIL;
   else if (token_type != T_IN)
   {print_msg (IN_ERR, token_buf);
    return_value = FAIL;}
   else
   {token_type = get_token();
switch (stype){
  case T_GET:
	 *repeat_label = 0;
  case T_WHILE:
  case T_FOR:
    log_exp(&expr_index,&expr_offset);
    if(expr_index == FAIL) return_value = FAIL;
    else if (symbol_table[expr_index].mode2 != T_PARALLEL)
     {print_msg (PAREXP_ERR, token_buf);
	   return_value = FAIL;}
    else if (symbol_table[expr_index].mode1 != T_LOGICAL)
     {print_msg (LOGEXP_ERR, token_buf);
	   return_value = FAIL;}
    else
/*       save_em(expr_index,expr_offset);    */
    break;
  case T_NEXT:
	 *repeat_label = 0;
    if (token_type != T_ID)
     {print_msg (ID_ERR, token_buf);
      return_value = FAIL;}
    else if ((expr_index = symtab_fetch(token_size, token_buf)) == FAIL)
     {print_msg (UNDEF_ERR, token_buf);
      return_value = FAIL;}
    else if (symbol_table[expr_index].mode2 != T_PARALLEL ||
             symbol_table[expr_index].mode1 != T_LOGICAL)
     {print_msg (LOGPAR_ERR, token_buf);
      return_value = FAIL;}
    else
     {token_type = get_token();
      if(subscript(expr_index,&arg1_index,&arg1_offset) == FAIL)
	         return_value = FAIL;
      expr_index=arg1_index;
      expr_offset=arg1_offset;}
    break;
   }
  if(return_value != FAIL)
      {switch(stype){
         case  T_GET:
            print_opcode ("get  ", stype, T_NONE, T_NONE);
				break;
         case  T_NEXT:
            print_opcode ("next ", stype, T_NONE, T_NONE);
				break;
         case  T_FOR:
            print_opcode ("for  ", stype, T_NONE, T_NONE);
				break;
         case  T_WHILE:
            print_opcode ("while", stype, T_NONE, T_NONE);
				break;
			default: printf("ERROR - illegal stype %d\n",stype);}
 	    print_args (off_exp1(expr_index,expr_offset),
		             off_exp2(invar_index,invar_offset),
		             end_label, repeat_label, NULL);
/*       do_it(invar_index,invar_offset);*/
       if((return_value = body()) != FAIL)
	     {if(token_type == T_ELSENANY)
			  {print_opcode("elsenany",T_ELSENANY,T_NONE,T_PARALLEL);
				print_args(else_label,end_label,NULL,NULL,NULL);
				token_type = get_token();
				return_value = body();}
			else
			  {print_opcode("label",T_LABEL,T_NONE,T_NONE);
				print_args(else_label,NULL,NULL,NULL,NULL);}
         if(stype == T_GET && token_type != T_ENDGET)
	        {print_msg (ENDGET_ERR, token_buf);
	         return_value = FAIL;}
	      else if (stype == T_NEXT && token_type != T_ENDNEXT)
	        {print_msg (ENDNEXT_ERR, token_buf);
	         return_value = FAIL;}
	      else if (stype == T_FOR && token_type != T_ENDFOR)
	        {print_msg (ENDFOR_ERR, token_buf);
	         return_value = FAIL;}
	      else if (stype == T_WHILE && token_type != T_ENDWHILE)
	        {print_msg (ENDWHILE_ERR, token_buf);
	         return_value = FAIL;}
         else
	        {token_type = get_token();
	         if (strcmp(token_buf, symbol_table[invar_index].tokpts) != 0)
	           {print_msg (INXUNM_ERR, token_buf);
		         return_value = FAIL;}
	         else
				  {switch(stype){
					 case T_GET:
			        print_opcode ("endget", T_ENDGET, T_NONE, T_NONE);
                 print_args (end_label, NULL, NULL, NULL, NULL);
                 break;
                case T_NEXT:
			        print_opcode ("endnext", T_ENDNEXT, T_NONE, T_NONE);
                 print_args (end_label, NULL, NULL, NULL, NULL);
                 break;
                case T_WHILE:
			        print_opcode ("endwhile", T_ENDWHILE, T_NONE, T_NONE);
                 print_args (end_label, repeat_label, NULL, NULL, NULL);
                 break;
                case T_FOR:
			        print_opcode ("endfor", T_ENDFOR, T_NONE, T_NONE);
                 print_args (off_exp1(invar_index,invar_offset), end_label,
									  repeat_label, NULL, NULL);
                 break;
			       default:
				     printf("ERROR - incorrect end statement for stype %d\n",stype);}
		         if((token_type = get_token()) == T_LSQU)
		          {while ((token_type = get_token()) != T_RSQU);
		           token_type = get_token();}
		         if(token_type != T_EOS)
		           {print_msg (EOS_ERR, token_buf);
		            return_value = FAIL;}
		         else
		           {token_type = get_token();
		            return_value = SUCCESS;}
   }}}}}

   if (debug == 2)
      printf("All() returning return_value %d\n", return_value);

   return(return_value);
}
