/* P1_ASSOCIATE 910415 */

#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "alloc.h"
#include "p1_common.h"
#include "p1_parser.h"
#include "p1_scanner.h"

/* ===========================================================================

   function name: associate
   input        : ---
   output	    : ---
   return value : return_value	- SUCCESS, if no error detected during parsing
				                     - FAIL, otherwise

   global vars  : token_type	- type of current token
		            token_buf	- current token string
		            symbol_table 	- symbol table data structure
		            debug		- debugging statement flag

   description  : This routine is parsing the associate statement

			associate par_var[$] PVAR_LIST with log_par_var;

			PVAR_LIST <-- , par_var[$] PVAR_LIST |
			             @

			pvar_var[$] is parallel varaible that should have been
				      declared in varaible declaration.

		   It will check the right syntax for the associate statement,
		   data type and parallel subscript "[$]". If an error is
		   detected during the parsing, a FAIL value will be returned,
		   otherwise a SUCCESS value will be returned.

=========================================================================== */

int associate()
{
   int return_value,bi_dx;

   if (debug == 2)
      printf("In associate_()...................................\n");

   return_value = SUCCESS;

   while (token_type == T_ASSOCIATE && return_value == SUCCESS)
     {token_type = get_token();
      if ((bi_dx = get_plist()) == FAIL)
         print_msg (ASSOC_ERR, NULL);
      else if ((return_value = pvar_list()) == FAIL)
        {print_msg(ASSOC_ERR, NULL);
          return_value = FAIL;}
      symbol_table[bi_dx].bi = return_value;
      return_value = SUCCESS;
	  }

   if ((token_type = get_token()) != T_EOS)
     {print_msg (EOS_ERR, token_buf);
      return_value = FAIL;}


   token_type = get_token();

   if (debug == 2)
      printf ("associate_() returning return_value : %d\n", return_value);

   return(return_value);

} /* end of associate_() */




/* ===========================================================================

   function name: pvar_list
   input	       : ---
   output	    : ---
   return value : return_value	- SUCCESS, if no error during parsing the
				                                variable list
				                     - FAIL, otherwise

   global vars  : token_type	- type of current token
		            token_buf	- current token string

   description  : This routine is parsing the syntax of parallel variables
		            and parallel subscript "[$]".

=========================================================================== */

int pvar_list()
{
   int return_value, id_index, sav_dx;

   if (debug == 2)
      printf("In pvar_list().................................\n");

   if (token_type == T_WITH)
     if ((sav_dx = get_pindex()) == FAIL)
        {print_msg(PARVAR_ERR, token_buf);
         return_value = FAIL;}
     else
		  return(sav_dx);
	else if (token_type == T_EOS)
        {printf("%d ERROR - illegal termination of associate statement\n",linenum);
			return_value = FAIL;}
   else
     {if (token_type == T_COMMA) token_type = get_token();
      if ((id_index = get_plist()) == FAIL)
        {print_msg (PARVAR_ERR, token_buf);
         return_value = FAIL;}
      else
        {if((return_value = pvar_list()) == FAIL)
			  {printf("%d ERROR - pvar_list failure\n",linenum);
				return(FAIL);}
			symbol_table[id_index].bi = return_value;}
     }

   if (debug == 2)
      printf ("pvar_list() returning return_value : %d\n");

   return(return_value);

} /* end of pvar_list() */




/* ===========================================================================

   function name: get_plist
   input     	 : ---
   output	    : ---
   return value : return_value	- SUCCESS, if no error is detected during
				                                parsing the varaible list
				                     - FAIL, otherwise

   global vars  : token_type 	- type of current token
		            token_buf	- current token string
		            symbol_table	- symbol table data structure
		            debug		- debugging statement flag

   description	 : This routine will parse the varaible list in the associate
		            statement. It will check the correct type, mode, occurance
		            and parallel subscript "[$]" of the varaible(s).

=========================================================================== */

int get_plist()
{
   int return_value;


   if (debug == 2)
      printf ("In get_plist()........................................\n");

   return_value = FAIL;

   if (token_type != T_ID)
     {print_msg (ID_ERR, token_buf);
      return_value = FAIL;}
   else if ((return_value = symtab_fetch (token_size, token_buf)) == FAIL)
     {print_msg (UNDEF_ERR, token_buf);
      return_value = FAIL;}
   else if (symbol_table[return_value].mode2 != T_PARALLEL)
     {print_msg(PARVAR_ERR, token_buf);
      return_value = FAIL;}

   if (check_parallel() == FALSE)
     {print_msg (SUBSCR_ERR, NULL);
      return_value = FAIL;}
   else
     token_type = get_token();

   if (debug == 2)
      printf("get_plist() returning return_value : %d, %s\n",
				  return_value,symbol_table[return_value].tokpts);

   return(return_value);

} /* end of get_plist() */




/* ===========================================================================

   function name: check_parallel
   input	       : ---
   output	    : ---
   return value : return_value	- TRUE, if no error is detected during parsing
				                             "[$]"
				                     - FALSE, if there is no parallel subscript
				                              "[$]" in the parallel varible
				                     - FAIL, if there is an error during parsing the
				                             parallel subscript "[$]"

   global vars  : token_type 	- type of current token
		            token_buf	- current token string
		            symbol_table	- symbol table data structure
		            debug		- debugging statement flag

   description	 : This routine is checking the occurance of the parallel
		            subscript "[$]" of the variable list in the associate
		            statement.

=========================================================================== */

int check_parallel()
{
  int return_value;

  if (debug == 2)
     printf ("In check_parallel()....................................\n");

  if ((token_type = get_token()) != T_LSQU)
     return_value = FALSE;
  else
    {if ((token_type = get_token()) != T_PAR_SUBSCRIPT)
        return_value = FAIL;
     else if ((token_type = get_token()) != T_RSQU)
        return_value = FAIL;
     else return_value = TRUE;
    }

  if (debug == 2)
     printf("check_parallel() returning return_value : %d\n");

  return(return_value);

} /* end of check_parallel() */



/* ===========================================================================

   function name: get_pindex
   input	       : ---
   output	    : ---
   return value : parvar_index	- index of the index parallel varaible in the
				                       symbol table
				                     - FAIL, otherwise

   global vars  : token_type 	- type of current token
		            token_buf	- current token string
		            symbol_table	- symbol table data structure.
		            debug		- debugging statement flag

   description	 : This routine is parsing the logical variable in the allocate
		            statement. It will check the data type and occurance in
		            the symbol table. It will also check the valid syntax for
		            parallel subscript "[$]".

=========================================================================== */

get_pindex()
{
   int return_value,
       parvar_index;

   if (debug == 2)
      printf ("In get_pindex().....................................\n");

   if ((token_type = get_token()) != T_ID)
     {print_msg (ID_ERR, token_buf);
      return_value = FAIL;}
   else if ((parvar_index = symtab_fetch(token_size, token_buf)) == FAIL)
     {print_msg (UNDEF_ERR, token_buf);
      return_value = FAIL;}
   else if (symbol_table[parvar_index].mode2 != T_PARALLEL)
     {print_msg (PARVAR_ERR, token_buf);
      return_value = FAIL;}
   else if ((token_type = get_token()) != T_LSQU)
     {print_msg(LSQU_ERR, token_buf);
      return_value = FAIL;}
   else if ((token_type = get_token()) != T_PAR_SUBSCRIPT)
     {print_msg (SUBSCR_ERR, token_buf);
      return_value = FAIL;}
   else if ((token_type = get_token()) != T_RSQU)
     {print_msg(RSQU_ERR, token_buf);
      return_value = FAIL;}
   else
      return_value = parvar_index;

   if (debug == 2)
      printf ("get_pindex() returning return_value : %d\n", return_value);

   return(return_value);

} /* end of get_pindex() */
