/* P1_DECLARE.C 910415 */

#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "alloc.h"
#include "p1_common.h"
#include "p1_parser.h"
#include "p1_scanner.h"
/* ===========================================================================

   function name: declare
   input	       : ---
   output       : ---
   return value : ret_value	- SUCCESS, if no errors detected
                        		- FAIL, otherwise

   global vars  : token_type	- type of token
                  token_buf	- token string
		            debug		- debugging statement flag

   description  : This routine is part of recursive descent parser and handles
		            the following productions from the grammar :

		     DECLARE  <-- int  MODE1 VAR VARLIST ; DECLARE |
		                  oct  MODE1 VAR VARLIST ; DECLARE |
		                  bin  MODE1 VAR VARLIST ; DECLARE |
		                  hex  MODE1 VAR VARLIST ; DECLARE |
		                  card MODE1 VAR VARLIST ; DECLARE |
                        char MODE1 VAR VARLIST ; DECLARE |
				            real MODE1 VAR VARLIST ; DECLARE |
				         logical MODE1 VAR VARLIST ; DECLARE |
				           index MODE1 VAR VARLIST ; DECLARE |
				              @

		     MODE     <-- scalar |
				            parallel

		 int is decimal
       '@' represents the empty production. MODE can be 'scalar'
		 or 'parallel'. The function VAR is called to process the
		 first identifier (which is required) in the declaration.
		 The function VARLIST is called to process the rest of the
		 identifiers. This routine returns FAIL if a syntax error
		 occurs in a declaration. SUCCESS is returned otherwise.
                 A special case for defvar declaration is it will not
	 	 generate intermediate code for the assembly code. But it
		 is still stored in the symbol table and acting as a normal
		 varible declaration.

=========================================================================== */

int declare()
{
   int ret_value;		/* identifier for the return value */
   int mode1;			/* type T_INT, T_REAL, T_INDEX, T_LOGICAL
				            T_CARD, T_HEX, T_BIN, T_OCT, T_CHAR */
   int mode2;			/* type T_SCALAR or T_PARALLEL */

   ret_value = SUCCESS;
  if (debug == 2)
      printf ("In declare()............................\n");

   mode1 = token_type;
   if (mode1 == T_INT || mode1 == T_REAL || mode1 == T_INDEX || mode1 == T_CARD
       || mode1 == T_LOGICAL || mode1 == T_OCT || mode1 == T_CHAR
       || mode1 == T_HEX     || mode1 == T_BIN || mode1 == T_LONGINT )
     {mode2 = get_token();
      if (mode2 == T_SCALAR || mode2 == T_PARALLEL || mode2 == T_CNST)
        {token_type = get_token();
         if ((ret_value = var(mode1, mode2)) != FAIL)
	        {if ((ret_value = varlist(mode1,mode2)) != FAIL)
	           {if (token_type == T_EOS)
	              {token_type = get_token();
		            ret_value = declare();}
   	         else
	              {print_msg(EOS_ERR,token_buf);
		            ret_value = FAIL;}
	           }
           }
        }
      else
        {print_msg(MODE_ERR, token_buf);
	      ret_value = FAIL;}
      }

  if (debug == 2)
      printf ("Declare returning %d\n", ret_value);
   return(ret_value);

} /* end of declare() */


/* ===========================================================================

   function name: var
   input	       : mode1		- type T_INT, T_REAL, T_INDEX, T_LOGICAL, T_OCT
                                       T_CARD, T_BIN, T_HEX, T_CARD
                  mode2		- type T_SCALAR or T_PARALLEL or T_CNST
   output	    : ---
   return value : ret_value	- SUCCESS, if no error detected
				                  - FAIL, otherwise

   global vars  : token_type	- type of current token
		            token_size	- size of current token
		            token_buf     - current token buffer
                  debug		- flag for debugging statement

   description	 : This routine is part of a recursive descent parser and
		            handles the following production from the grammar:

			VAR <-- identifier WIDTH DIMENSION IOSPEC

		  If the current token is not an identifier, it is an error.
		  Otherwise the functions WIDTH, DIMENSION and IOSPEC are called
		  to check if a width or dimensions were specified. If no
		  errors are detected, the identifier is inserted in the
		  symbol table, the intermediate code is printed, and
		  SUCCESS is returned. Otherwise the return value is FAIL.

=========================================================================== */

int var(mode1, mode2)
int mode1, mode2;
{
   int ret_value;		/* identifier for return value */
   int id_size;		/* copy of identifier token size */
   int arg1_index;	/* symbol table index for identifier */
   int i_width;		/* integer for the bit field width */
   int i_expt;       /* integer for the exponent width - real only */
   int i_signif;     /* integer for the mantisa  width - real only */
   int i_total_size;	/* integer for total array size */
   int i;			   /* loop counter */
   int i_num_dim;		/* integer for number of dimensions */
   int i_dim_size[MAX_NUM_DIM]; /* integer array for size of each dimension */
   int specdx;
   char c_mode[10];  /* character form of mode1+mode2 */
   char c_width[6];	/* character buffer for the bit field width */
   char c_expt[6];   /* character buffer for the exponent field width */
   char c_signif[6]; /* character buffer for the mantisa field width */
   char c_total_size[10];	/* character buffer for total array size */
   char id_buf[TOKEN_SIZE];/* copy of name of identifier token */

   if (debug == 2)
      printf ("In var()...........................\n");

   ret_value = SUCCESS;

   for (i = 0; i < MAX_NUM_DIM; i++)
      i_dim_size[i] = 0;

   if (token_type != T_ID)
     {print_msg(ID_ERR, token_buf);
      ret_value = FAIL;}
   else
     {strcpy(id_buf, token_buf);
      id_size = token_size;
      token_type = get_token();
      if ((ret_value = width(mode1,mode2,&i_width,&i_expt,&i_signif)) != FAIL)
         {ret_value = dimension(mode2, &i_total_size, i_dim_size, &i_num_dim);
			 specdx = iospec();
          if(ret_value != FAIL)
           {pc_itoa(i_width, c_width);
	         pc_itoa(i_expt, c_expt);
	         pc_itoa(i_signif, c_signif);
	         pc_itoa(i_total_size, c_total_size);
	         pc_itoa(mode1*SIZE_MODE2+mode2,c_mode);
/* check that any defvars are declared in the proper order */
/* if defineing ida, idb must already be defined */
	         for (i=0; i<defvar_cnt && ret_value != FAIL; i++)
	            if(strcmp(defvar_table[i].ida,id_buf) == 0)
		            if(symtab_fetch(strlen(defvar_table[i].idb),
				         defvar_table[i].idb) == FAIL)
		              {printf("ERROR - must define %s of defvar before %s\n",
			                         defvar_table[i].idb,defvar_table[i].ida);
		               ret_value = FAIL;}
	         if(ret_value != FAIL)
	           {if (mode1 != T_INDEX)  /* all non INDEX declarations */
	              arg1_index = symtab_insert(T_ID, mode1, mode2, id_size, id_buf,
                                 i_dim_size, i_width,i_expt,i_signif, i_num_dim);
	            else   /* scalar and  parallel index */
                 arg1_index = symtab_insert(T_INDEX, T_LOGICAL, mode2, id_size,
                                  id_buf, i_dim_size, i_width,i_expt,i_signif, i_num_dim);
               if (arg1_index != FAIL)
	              {if (mode2 == T_SCALAR)
                    {if (module_flag == T_MAIN)
                       {install_scalar(id_buf, c_total_size, "0",
				                       c_mode,c_width,c_expt,c_signif,FALSE,specdx);
		                  print_opcode("entry_", T_ENTRY, T_NONE, T_NONE);
		                  print_args(id_buf, NULL, NULL, NULL, NULL);}
		               else if (module_flag == T_SUBROUTINE)
                        {install_scalar(id_buf, c_total_size, "0",
      	                          c_mode,c_width,c_expt,c_signif,FALSE,specdx);
                         if(code_type != EMULATE)
									{print_opcode("extern_",T_EXTERN, T_NONE, T_NONE);
									 print_args(id_buf, NULL,NULL,NULL,NULL);}
                        }
	                 }
                  else if(code_type != EMULATE || module_flag != T_SUBROUTINE)
		              {if(mode1 == T_INDEX)
		                 print_opcode ("decl_", T_DECLARE, T_LOGICAL, mode2);
		     		      else
		                 print_opcode ("decl_", T_DECLARE, mode1, mode2);
   	               print_args6(id_buf, c_total_size, c_width, c_expt, c_signif,
											symbol_table[specdx].tokpts);
                    }
 					  if(specdx != FAIL)
								iospec_insert(symtab_fetch(strlen(id_buf),id_buf),specdx);
                 }
	            else
	              ret_value = FAIL;
	           }
	        }
        }
     }

    if (debug == 2)
       printf ("Var() returning ret_value = %d\n", ret_value);

    return(ret_value);

} /* end of var() */


/* ===========================================================================

   function name : varlist
   input 	     : mode1		- type T_INT, T_REAL, T_INDEX, or T_LOGICAL
                                       T_CARD
		             mode2		- type T_SCALAR or T_PARALLEL or T_CNST
   output        : ---
   return value  : ret_value	- SUCCESS, if no errors detected
	          		            - FAIL, otherwise

   global vars   : token_type    - type of token
		             debug		- debugging statement flag

   description   : This routine is part of a recursive descent parser and
		             handles the following production from the grammar :

		     VARLIST <-- , VAR VARLIST
				        |  @

		  '@' represents the empty production. This routine is used
		  to process the remaining IDs in a single declaration. If
	 	  any are found, VAR is called to process them and insert
		  them in the symbol table. The function returns FAIL if a
		  syntax error occurs; SUCCESS is returned otherwise.

=========================================================================== */

int varlist(mode1, mode2)
int mode1, mode2;
{
   int ret_value;		/* value returned by function */

   ret_value = SUCCESS;

   if (debug == 2)
      printf ("In varlist().........................\n");

   if (token_type == T_COMMA) token_type = get_token();
   if (token_type == T_ID)
     {if ((ret_value = var(mode1, mode2)) != FAIL)
         ret_value = varlist(mode1, mode2);}

   if (debug == 2)
      printf ("Varlist is returning %d\n", ret_value);

   return(ret_value);

} /* end of varlist() */


/* ===========================================================================

   function name  : width
   input	         : mode1		- type T_INT, T_REAL, T_INDEX, T_LOGICAL
                                       T_CARD
                    mode2		- type T_SCALAR or T_PARALLEL or T_CNST

   output	      : i_width	- integer pointer to bit field width

   return value   : ret_value	- SUCCESS, if no error detected
				                  - FAIL, otherwise

   global vars    : token_type	- type of current token
		              token_buf     - current token buffer
                    debug		- flag for debugging statement

   description	   : This routine is part of a recursive descent parser and
		              handles the following production from the grammar:

			WIDTH <-- : decimal_constant |
                   : decimal_constant : decimal_constant |
				       @

		  '@' represents the empty production. If the current
		  token is a ':', and the next token is a decimal integer
		  constant, it is used as the field width. If a colon is
		  not seen, a default width is used. FAIL is returned if
		  any errors are detected; otherwise SUCCESS is returned.

=========================================================================== */

int width(mode1, mode2, i_width, i_expt, i_signif)
int mode1, mode2;
int *i_width,*i_expt,*i_signif;
{
   int ret_value;		/* identifier for return value */
   int d_width;

   if (debug == 2)
      printf ("In width()...........................\n");

   ret_value = SUCCESS;

   *i_expt = 0;    /* default expt and signif lengths */
   *i_signif = 0;

   switch (mode1)
     {case T_INT   : *i_width = *i_signif = g_width(W_INT,M_INT);
	                  break;
      case T_LONGINT: *i_width = *i_signif = g_width(W_LONGINT,M_INT);
	                  break;
      case T_CARD  : *i_width = *i_signif = g_width(W_CARD,M_CARD);
	                  break;
      case T_BIN   : *i_width = *i_signif = g_width(W_BIN,M_BIN);
	                  break;
      case T_HEX   : *i_width = *i_signif = g_width(W_HEX,M_HEX);
	                  break;
      case T_OCT   : *i_width = *i_signif = g_width(W_OCT,M_OCT);
   	               break;
      case T_REAL  : *i_expt = g_width(W_EXPT,M_EXPT);
	                  *i_signif = g_width(W_SIGNIF,M_SIGNIF);
                     *i_width = *i_expt + *i_signif;
   	               break;
      case T_CHAR  : *i_width = *i_signif = g_width(W_CHAR,M_CHAR);
   	               break;
      case T_INDEX : if (mode2 == T_SCALAR) d_width = W_INT;
		               else d_width = W_LOGICAL;
	                  *i_width = *i_signif = g_width(d_width,M_INDEX);
                     break;
      case T_LOGICAL: *i_width = *i_signif = g_width(W_LOGICAL,M_LOGICAL);
		               break;
     }

    if (debug == 2)
       printf ("Width() returning ret_value = %d\n", ret_value);

    return(ret_value);

} /* end of width() */


/* g(et)_width looks for a colon to signify the presence of a user
   specified width value. If present it converts it to int and checks
   to see if it is greater than m(inimum)_width. It returns the larger
   of the converted value and m_width.  If not present, it returns the
   default value.  It also checks to see that the size does not exceed
   the size of */

g_width(d_width,m_width)
int d_width,m_width;
{  int ret_value;

  if (token_type == T_COLON)
    {if ((token_type = get_token()) != T_DEC_CONST && token_type != T_ID)
       {print_msg(WIDTH_ERR, token_buf);
	     ret_value = FAIL;}
     else
       {if (token_type == T_DEC_CONST) ret_value = atoi(token_buf);
        else if ((ret_value = symtab_fetch(token_size,token_buf))== FAIL)
	        print_msg(UNDEF_ERR, token_buf);
	     else if (symbol_table[ret_value].mode1 != T_INT &&
	         symbol_table[ret_value].mode2 != T_SCALAR)
	       {print_msg(WIDTH_ERR, token_buf); ret_value = FAIL;}
        else if((ret_value = define_fetch(token_buf))== FAIL)
 	        print_msg(UNDEF_ERR, token_buf);
	     else ret_value = atoi(scalar_table[ret_value].arg3); }

     if (ret_value != FAIL)
       {if (ret_value < m_width)
  	       {print_msg(WIDTH_ERR, token_buf);
           ret_value = m_width;}
	else if (ret_value > SC_TEMP_SIZE)
               printf("%d WARNING the maximum emulator variable size is %d\n",linenum,SC_TEMP_SIZE);
	token_type = get_token();
       }
     }
   else ret_value = d_width;

  return(ret_value);
}



/* ===========================================================================

   function name: dimension
   input	       : mode2			   - type T_SCALAR or T_PARALLEL
   output	    : i_total_size	- pointer to total array size
		            i_num_dim		- pointer to number of dimensions
		            i_dim_siz		- array of dimension sizes
   return value : ret_value		- SUCCESS if no errors detected
					                  - FAIL, otherwise

   global vars  : token_type		- type of current token
		            token_buf		- current token buffer
		            debug			- flag for debugging statement

   description	 : This routine is part of a recursive descent parser and
		            handles the following production from the grammar:

		 	DIMENSION <-- [ $ DIMLIST ] |
				           [ decimal_constant DIMLIST ] |
				           @

		  '@' represents the empty production. Dimensions are
		  optional, unless the identifier was parallel. In that
		  case, at least one dimension is required. The first
		  dimension of a parallel identifier must be '$'. The
		  first dimension of a scalar identifier must be a
		  decimal integer constant. The function DIMLIST is
		  called to process the rest of the dimensions, if any.
		  This function calculates the total number of array
		  elements, the number of dimensions, and the size of
		  each dimension for access by the calling routine. The
		  return value is SUCCESS, if no errors are detected and
		  FAIL otherwise.

=========================================================================== */

int dimension(mode2, i_total_size, i_dim_size, i_num_dim)
int mode2;
int *i_total_size, *i_num_dim;
int i_dim_size[];
{
	int size;				/* size of a dimension */
	int ret_value;			/* return value of function */


	/* initialize local variables - MCS - 08/06/2002 */
	size = 0;
	ret_value = SUCCESS;

	if (debug == 2)
		printf("In dimension()...................\n");

	*i_num_dim = 0;
	*i_total_size = 1;

	if (token_type == T_LSQU)
	{
		token_type = get_token();
		if ((mode2 == T_PARALLEL && token_type != T_PAR_SUBSCRIPT) || 
			(mode2 == T_SCALAR && (token_type != T_DEC_CONST && token_type != T_ID)))
		{
			print_msg(DIM_ERR, token_buf);
			ret_value = FAIL;
		}
		else
        {
			if (mode2 == T_SCALAR)
	        {
				if (token_type == T_DEC_CONST)
					size = atoi(token_buf);
				else
				{
					if ((ret_value = symtab_fetch(token_size, token_buf)) == FAIL)
						print_msg(UNDEF_ERR, token_buf);
					else if (symbol_table[ret_value].mode1 != T_INT && symbol_table[ret_value].mode2 != T_SCALAR)
					{
						print_msg(DIM_ERR, token_buf);
						ret_value = FAIL;
					}
					else if ((ret_value = define_fetch(token_buf)) == FAIL)
						print_msg(UNDEF_ERR, token_buf);
					else
						size = atoi(scalar_table[ret_value].arg3);
				}

				if (ret_value != FAIL)
				{
					i_dim_size[*i_num_dim] = size;
					*i_total_size = *i_total_size * size;
				}
			}

			if (ret_value != FAIL)
			{
				*i_num_dim = *i_num_dim + 1;
				token_type = get_token();
				if((ret_value = dimlist(i_total_size,i_dim_size,i_num_dim)) != FAIL)
				{
					if (token_type != T_RSQU)
					{
						print_msg(RSQU_ERR, token_buf);
						ret_value = FAIL;
					}
					else
						token_type = get_token();
				}
			}
		}
	}
	else if(mode2 == T_PARALLEL)
	{
		print_msg(LSQU_ERR, token_buf);
		ret_value = FAIL;
	}

	if (debug == 2)
		printf ("Dimension() returning ret_value %d\n", ret_value);

	return(ret_value);

} /* end of dimension() */



/* ===========================================================================

   function name: dimlist
   input	       : ---
   output	    : i_total_size	- pointer to total array size
		            i_dim_size		- array of dimension sizes
		            i_num_dim		- pointer to number of dimensions
   return value : ret_value		- SUCCESS if no errors
					                  - FAIL, otherwise

   global vars  : token_type		- type of current token
		            token_buf		- current token buffer
		            debug			- flag for debugging statements

   description	 : This routine is part of a recursive descent parser and
		            handles the following production from the grammar:

		 	DIMLIST <-- , decimal_constant |
				         @

		  '@' represents the empty production. If a 'comma' is
		  found, the next token must be a decimal integer constant
		  and is used as the next dimension. The number of dimensions
		  may not exceed MAX_NUM_DIM.

=========================================================================== */

int dimlist(i_total_size, i_dim_size, i_num_dim)
int *i_total_size, *i_num_dim;
int i_dim_size[];
{
   int ret_value;		/* return value of function */
   int size;			/* size of 1 dimension */

   if (debug == 2)
      printf("In dimlist().............................\n");

   ret_value = SUCCESS;

   if (*i_num_dim < MAX_NUM_DIM && token_type == T_COMMA)
     {if ((token_type = get_token()) != T_DEC_CONST &&
           token_type != T_ID)
        {print_msg(DIM_ERR, token_buf);
	      ret_value = FAIL;}
      else
        {if (token_type == T_DEC_CONST)
            size = atoi(token_buf);
         else if ((ret_value = symtab_fetch(token_size, token_buf)) == FAIL)
		      print_msg(UNDEF_ERR, token_buf);
	      else if (symbol_table[ret_value].mode1 != T_INT &&
		            symbol_table[ret_value].mode2 != T_SCALAR)
	        {print_msg(DIM_ERR, token_buf);
	         ret_value = FAIL;}
	      else if((ret_value = define_fetch(token_buf)) == FAIL)
 	         print_msg(UNDEF_ERR, token_buf);
	      else
	        size = atoi(scalar_table[ret_value].arg3);
        }

      if (ret_value != FAIL)
        {i_dim_size[*i_num_dim] = size;
	      *i_total_size = *i_total_size * size;
	      *i_num_dim = *i_num_dim + 1;
	      token_type = get_token();
 	      ret_value = dimlist(i_total_size, i_dim_size, i_num_dim);
        }
     }

   if (debug == 2)
      printf("Dimlist() returning ret_value %d\n", ret_value);

   return(ret_value);

} /* end of dimlist() */

iospec()
{  int result_index;

	if(token_type == T_STRING)
	  {result_index = get_const(token_buf,token_type);
      token_type=get_token();}
	else
		result_index = shapedx;
	return(result_index);
}
