/* 910415  P1_IF.C */

#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "alloc.h"
#include "p1_common.h"
#include "p1_parser.h"
#include "p1_scanner.h"

/* ===========================================================================

   function name : if_
   input	        : ---
   output	     : ---
   return value  : return_value	- SUCCESS, if no errors
				                     - FAIL, otherwise

   global vars   : symbol_table	- array containing symbol table
		             if_count	- counter for making 'if' statement labels
		             token_type	- type of current token
		             token_buf	- character string for current token
		             debug		- flag for debugging statements

   description	  : This routine is part of a recursive descent parser and
		             handles the following production from the grammar:

			IF_ <-- LOG_EXP then BODY MORE_IF endif ;

		  The logical expression may be scalar or parallel. It
		  determines the code to be generated for the remainder
		  of the statement. If the expression result is parallel,
		  the 'then' portion is executed for all non-zero responders
		  and the 'else' portion (if any) is executed for all zero
		  responders. If the expression result is scalar, the 'then'
		  portion is executed if the expression is non-zero and
		  the 'else' portion (if any) is executed if the expression
		  is zero. The 'then' and 'endif' keywords are required.

=========================================================================== */

int if_()
{
   int return_value;		/* return value of function */
   int expr_index;		/* symbol table index of expression result */
   int expr_offset;
   int log_mode;		   /* logical expression result mode: SCALAR or
				               PARALLEL */
   char *start_label;	/* label to print at start of 'if' */
   char *end_label;		/* label to print at end of 'if' */
   char *repeat_label;	/* label to print at middle of 'if' */

   if (debug == 2)
      printf ("In if_()........................................\n");

   start_label = make_label ("IF$S", ++if_count);
   end_label = make_label ("IF$E", if_count);
   repeat_label = make_label("ELS$", if_count);

   print_opcode ("begif", T_BEGIF, T_NONE, T_NONE);
   print_args (start_label, NULL, NULL, NULL, NULL);

   log_exp(&expr_index,&expr_offset);
   if (expr_index == FAIL) return_value = FAIL;
   else if (symbol_table[expr_index].mode1 != T_LOGICAL)
     {print_msg (LOGEXP_ERR, token_buf);
      return_value = FAIL;}
   else
     {save_em(expr_index,expr_offset);
      log_mode = T_SCALAR;
	if (symbol_table[expr_index].mode2 == T_PARALLEL)
         log_mode = T_PARALLEL;
      else if (symbol_table[expr_index].mode2 == T_SCALAR
       /*|| symbol_table[expr_index].mode2 == T_SCALAR_REG */)
         log_mode = T_SCALAR;
      if (token_type != T_THEN)
        {print_msg(THEN_ERR, token_buf);
         return_value = FAIL;}
      else
	{token_type = get_token();
         print_opcode ("if_", T_IF, T_NONE, log_mode);
 	      print_args (off_exp1(expr_index,expr_offset),
							repeat_label, NULL, NULL, NULL);
         if ((return_value = body()) != FAIL)
	        {if (log_mode == T_PARALLEL)
	           if (token_type == T_ELSENANY)
		{/* mode1 T_IFNANY flags the T_ELSENANY as in a parallel if */
		 print_opcode("elsenany_",T_ELSENANY, T_IFNANY, T_PARALLEL);
                 print_args(repeat_label, end_label, NULL,NULL,NULL);}
	           else
		          {print_opcode("label", T_LABEL, T_NONE, T_NONE);
                 print_args(repeat_label, NULL, NULL, NULL, NULL);
		           if(token_type != T_ELSE) goto nobody;
                 print_opcode("else_", T_ELSE, T_NONE, T_PARALLEL);
                 print_args(end_label, NULL, NULL, NULL, NULL);
                }
            else /* log_mode == T_SCALAR */
              {print_opcode("else_", T_ELSE, T_NONE, T_SCALAR);
               print_args(end_label, repeat_label, NULL, NULL, NULL);
	            if(token_type != T_ELSE) goto nobody;
              }
	         token_type = get_token();
            return_value = body();   /* else, elsenany  body */
nobody:     if(return_value != FAIL)
              {if (token_type != T_ENDIF)
	              {print_msg (ENDIF_ERR, token_buf);
	               return_value = FAIL;}
	            else
	              {print_opcode ("endif_", T_ENDIF, T_NONE, log_mode);
		            print_args (end_label, NULL, NULL, NULL, NULL);
		            if ((token_type = get_token()) != T_EOS)
		              {print_msg (EOS_ERR, token_buf);
		               return_value = FAIL;}
		            else
		              {token_type = get_token();
		               return_value = SUCCESS;}
	              }
	           }
           }
        }
     }


   if (debug == 2)
      printf("If() returning return_value %d\n", return_value);

   return(return_value);
}




/* ===========================================================================

   function name : more_if
   input	        : label		- label to print at end of 'if'
		             log_mode	- mode of 'if' (PARALLEL or SCALAR)
   output	     : ---
   return value  : ret_value	- SUCCESS, if no errors
				                  - FAIL, otherwise

   global vars   : token_type	- type of current token
		             debug      - flag for debugging statements

   description	  : This routine is part of a recursive descent parser and
		             handles the following production from the grammar:

			MORE_IF <-- else BODY |
				    @

		  '@' represents the empty production. If no keyword
		  'else' is found, SUCCESS is returned. Otherwise, the
		  functions returns the value returned by a call to
		  'body'.

=========================================================================== */

int more_if(label, log_mode)
char *label;
int log_mode;
{
  int ret_value;		/* return value: SUCCESS or FAIL */

  if (debug == 2)
      printf ("In more_if()..................................\n");

  if (token_type != T_ELSE)
      ret_value = SUCCESS;
  else
    {if (log_mode == T_PARALLEL)
       {print_opcode("else_", T_ELSE, T_NONE, T_PARALLEL);
        print_args(label, NULL, NULL, NULL, NULL);}
     token_type = get_token();
     ret_value = body();
    }

  if (debug == 2)
     printf("More_if() returning ret_value %d\n",ret_value);

  return(ret_value);

} /* end of more_if() */


