/* 910415  P1_LOOP.C */

#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "alloc.h"
#include "p1_scanner.h"
#include "p1_common.h"
#include "p1_parser.h"

/* ===========================================================================

   function name : first_
   input	        : ---
   output	     : ---
   return value  : ret_value	- SUCCESS if no error detected in the loop
				                  - FAIL, otherwise

   global vars   : token_type	- type of current token
		             token_buf	- string of current token
		             debug		- flag for debugging statements

   description	  : This routine is part of the recursive decent parser and
		             handles the following production from the grammar:

		    FIRST_  <-- FIRST_BODY loop LOOP_

		  FIRST_BODY represents the initialization portion of the loop
		  and could be empty. The keyword 'loop' is required. The
		  LOOP_ function will handle code generation for the loop.

=========================================================================== */

int first_()
{
   int ret_value;		/* variable for return value */

   if (debug == 2)
      printf("In first_()..........................\n");

   if ((ret_value = first_body()) != FAIL)
     {if (token_type != T_LOOP)
        {print_msg(LOOP_ERR, token_buf);
         ret_value = FAIL;}
      else
        {token_type = get_token();
	      ret_value = loop_();}
     }
   if (debug == 2)
      printf("First() returning ret_value = %d\n", ret_value);

   return(ret_value);
}

/* ===========================================================================

   function name : loop_
   input	        : ---
   output	     : ---
   return value  : ret_value	- SUCCESS, if no errors
				                  - FAIL, otherwise

   global vars   : loop_count	- counter for making 'loop' statement labels
		             symbol_table	- array containing symbol table
		             debug		- flag for debugging statements
		             token_type	- type of current token
		             token_buf	- character string for current token

   description	  : This routine is part of a recursive descent parser and
		             handles the following production from the grammar:

			LOOP_ <-- BODY until LOG_EXP BODY MORE_LOOP endloop ;

		  Any of the calls to BODY may be empty. The keyword 'until'
		  is required and the expression which follows it must give
		  a logical scalar result. The call to MORE_LOOP will handle
		  optional multiple loop exit conditions. The keyword 'endloop'
		  is required. All of the statements in each particular BODY
		  are executed until the logical expression evaluates to
		  non-zero (TRUE). The loop is terminated at that point and
		  any other statements following the expression are not
		  executed.

=========================================================================== */

int loop_()
{
   int ret_value;	      /* return value of function */
   char *start_label;	/* label to print at start of 'loop' */
   char *end_label;	   /* label to print at end of 'loop' */
   int expr_index;	   /* symbol table index of expression result */
   int expr_offset;
   int nany_flg;        /* flag used to signify a nany keyword 880309 jlp*/

   if (debug == 2)
      printf("In loop_()..................................\n");

   start_label = make_label("LO$S", ++loop_count);
   end_label = make_label("LO$E", loop_count);

   print_opcode("loop   ", T_LOOP, T_NONE, T_NONE);
   print_args(start_label, NULL, NULL, NULL, NULL);

   if ((ret_value = body()) != FAIL)
     {if (token_type != T_UNTIL)
        {ret_value = FAIL;
         print_msg(UNTIL_ERR, token_buf);}
      else
        {nany_flg = FALSE;
         token_type = get_token();
	      if(token_type == T_NANY)
	        {token_type = get_token();
            nany_flg = TRUE;}
         if (log_exp(&expr_index,&expr_offset) == FAIL) ret_value = FAIL;
         else if (symbol_table[expr_index].mode1 != T_LOGICAL)
           {print_msg (LOGEXP_ERR, NULL);
            ret_value = FAIL;}
	      else if (symbol_table[expr_index].mode2 == T_SCALAR &&
						nany_flg == TRUE)
	        {printf("%4d ERROR - nany requires a parallel exression\n",linenum);
	         ret_value = FAIL;}
	      else
	        {save_em(expr_index,expr_offset);
			   if(nany_flg == TRUE)
				   print_opcode("until_",T_UNTIL,T_IFNANY,symbol_table[expr_index].mode2);
            else
				   print_opcode("until_",T_UNTIL,T_NONE,symbol_table[expr_index].mode2);
            print_args(off_exp1(expr_index,expr_offset), end_label,
							     NULL,NULL, NULL);
            if ((ret_value = body()) != FAIL)
              {if ((ret_value = more_loop(end_label)) != FAIL)
	              {if (token_type != T_ENDLOOP)
                    {print_msg(ENDLOOP_ERR, token_buf);
                     ret_value = FAIL;}
                  else if ((token_type = get_token()) != T_EOS)
                    {print_msg(EOS_ERR, token_buf);
                     ret_value = FAIL;}
                  else
                    {print_opcode("endloop", T_ENDLOOP, T_NONE, T_NONE);
		               print_args(end_label, start_label, NULL, NULL, NULL);
                     token_type = get_token();}
	              }
              }
           }
        }
     }

   if (debug == 2)
      printf ("Loop_() returning ret_value %d\n", ret_value);

   return (ret_value);

} /* end of loop_() */


/* ===========================================================================

   function name : more_loop
   input	        : end_label	- label to print at end of 'loop' statement
   output	     : ---
   return value  : ret_value	- SUCCESS, if no errors
				                  - FAIL, otherwise

   global vars   : token_type 	- type of current token
		             symbol_table	- array containing symbol table
		             debug		- flag for debugging statements

   description	  : This routine is part of a recursive descent parser and
		             handles the following production from the grammar:

		  	MORE_LOOP <-- until LOG_EXP BODY MORE_LOOP |
				      @

		  '@' represents the empty production. If the keyword
		  'until' is not found, the function returns SUCCESS.
		  Otherwise, a logical scalar expression must follow.
		  The call to BODY can be empty. The function is called
		  recursively to allow multiple exit conditions for
		  a loop.

=========================================================================== */

int more_loop(end_label)
char *end_label;
{
   int ret_value;	   /* value returned by function */
   int expr_index;	/* symbol table index of expression result */
   int expr_offset;
   int nany_flg;

   if (debug == 2)
      printf("In more_loop()................................\n");

   if (token_type != T_UNTIL)
      ret_value = SUCCESS;
   else
     {nany_flg = FALSE;
      token_type = get_token();
	   if(token_type == T_NANY)
	     {token_type = get_token();
         nany_flg = TRUE;}
         if (log_exp(&expr_index,&expr_offset) == FAIL) ret_value = FAIL;
         else if (symbol_table[expr_index].mode1 != T_LOGICAL)
           {print_msg (LOGEXP_ERR, NULL);
            ret_value =FAIL;}
	      else if(symbol_table[expr_index].mode2 == T_SCALAR &&
					  nany_flg == TRUE)
	        {print_msg_m1(NANY_ERR, NULL);
	         ret_value = FAIL;}
	      else
	        {if(nany_flg == TRUE)
				  print_opcode("until", T_UNTIL, T_IFNANY,symbol_table[expr_index].mode2);
            else
	           print_opcode("until", T_UNTIL, T_NONE,symbol_table[expr_index].mode2);
            print_args(off_exp1(expr_index,expr_offset), end_label,
							  NULL, NULL, NULL);
	         if ((ret_value = body()) != FAIL)
               ret_value = more_loop(end_label);}
        }

   if (debug == 2)
      printf("More_loop() returning ret_value %d\n", ret_value);

   return(ret_value);
}

/* ===========================================================================

   function name : first_body
   input         : ---
   output	     : ---
   return value  : ret_value	- SUCCESS, if no errors
				                  - FAIL, otherwise

   global vars   : debug		- flag for debugging statements
		             token_type	- type of current token
		             sca_temp	- symbol table index of current scalar temp
		             ipar_temp	- symbol table index of current int/par temp
		             rpar_temp	- symbol table index of current real/par temp
		             lpar_temp	- symbol table index of current log/par temp

   description	  : This routine is part of a recursive descent parser and
		             handles the following production from the grammar:

			FIRST_BODY <-- for FOR_ FIRST_BODY           |
				            next NEXT_ FIRST_BODY         |
				            get GET_ FIRST_BODY           |
				            if IF_ FIRST_BODY             |
				            any ANY_ FIRST_BODY           |
				            while WHILE_ FIRST_BODY       |
      				      asmcode ASSEMBLE_ FIRST_BODY  |
				            allocate ALLOCATE_ FIRST_BODY |
				            call CALL_ BODY               |
				            return RETURN_ BODY           |
				            setscope SETSCOPE_ BODY       |
				            identifier ASSIGN FIRST_BODY  |
				            @

		  No code is generated. If none of the specified tokens is
		  found, the function returns SUCCESS. If one of the tokens
		  is found and no errors are detected in subsequent function
		  calls, the function is called recursively to process the
		  next statement, until either an error occurs or no more
		  statements are found. This function is a modified form
		  of BODY which does not allow the keywords 'first' or
		  'loop' to be accepted. Its purpose is to prevent loop
		  statements from occurring in the initialization portion
		  of another loop statement.

=========================================================================== */

int first_body()
{
   int return_value;		/* return value of function */
   int done;			   /* recursion control variable */
   char num[13];

   if (debug == 2)
      printf ("In first_body(), token_type=%x\n",token_type);

   return_value = SUCCESS;
   done = FALSE;
   sca_temp = sca_temp_start;
   real_sca_temp = REAL_SCA_TEMP_START;
   ipar_temp = INT_PAR_TEMP_START;
   rpar_temp = REAL_PAR_TEMP_START;
   lpar_temp = log_par_temp_start;
   g_offset = g_offset_start;

   pc_itoa(linenum,num);
   print_opcode ("beg_of_stmt ", T_BOS, T_NONE, T_NONE);
   print_args (num,"0",NULL,NULL,NULL);

   if (token_type == T_FOR)
     {token_type = get_token();
      return_value = all_(T_FOR);}
   else if (token_type == T_GET)
     {token_type = get_token();
      return_value = all_(T_GET);}
   else if (token_type == T_ANY)
     {token_type = get_token();
      return_value = any_();}
   else if (token_type == T_NEXT)
     {token_type = get_token();
      return_value = all_(T_NEXT);}
   else if (token_type == T_IF)
     {token_type = get_token();
      return_value = if_();}
   else if (token_type == T_WHILE)
     {token_type = get_token();
      return_value = all_(T_WHILE);}
   else if (token_type == T_ALLOCATE)
     {token_type = get_token();
      return_value = allocate_();}
   else if (token_type == T_RELEASE)
     {token_type = get_token();
      return_value = release_();}
   else if (token_type == T_CALL)
     {token_type = get_token();
      return_value = call_();}
   else if (token_type == T_RETURN)
     {token_type = get_token();
      return_value = return_();}
   else if (token_type == T_STOP)
     {token_type = get_token();
      return_value = stop_();}
   else if (token_type == T_ASSEMBLE)
      return_value = assemble_();
   else if (token_type == T_ID)
      return_value = assign();
   else if (token_type == T_MSG)
      return_value = msgin_();
   else if (token_type == T_PRINT)
      return_value = print_();
   else
      done = TRUE;

   if (return_value == SUCCESS && !done)
      return_value = first_body();

   if (debug == 2)
      printf ("First_body() returning return_value %d\n", return_value);

   return(return_value);
}
