#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "alloc.h"
#include "p1_common.h"
#include "p1_parser.h"
#include "p1_scanner.h"

/* ===========================================================================

   function name: stackwhile_
   input	: ---
   output	: ---
   return value : return_value	- SUCCESS, if no errors
			 	- FAIL, otherwise

   global vars  : symbol_table	- array containing symbol table
		  stackwhile_count	- counter for making 'while' statement labels
		  token_type	- type of current token
		  token_buf	- character string for current token
		  debug		- flag for debugging statements

   description	: This routine is part of a recursive descent parser and
		  handles the following production from the grammar:

			STACKWHILE_ <-- index in LOG_EXP BODY endwhile index ;

		  The index variable must be an identifier that has been
		  declared as an 'index parallel' identifier. The expression
		  must evaluate to a logical parallel result. The body
		  of the 'while' statement can be empty. The keywords 'in'
		  and 'endwhile' are required. The index variable must be
		  repeated at the end of the 'while' loop. The expression
		  is evaluated at the beginning of each pass of the loop.
		  and the result is stored in the index variable. The loop
		  is executed as long as there are any non-zero responses.
		  The index variable may not be modified by the user during
		  the loop execution, but may be used to index a parallel
		  variable. Loop termination occurs as a result of the user
		  modifying some variable of the controlling expression
		  within the body of the loop.

=========================================================================== */

int stackwhile_()
{
   int return_value;		/* return value of function */
   int invar_index;		/* symbol table index of index variable */
   int invar_offset;
   int expr_index;		/* symbol table index of expression result */
   int expr_offset;
   char *end_label;		/* label to print at end of 'whilestack' */
   char *repeat_label;		/* label to print at top of 'whilestack' */
   char *recurse_label;		/* label to print at recurse of 'whilestack' */

   if (debug == 2)
      printf ("In stackwhile_()........................................\n");

   stack_end = end_label = make_label ("WI$E", ++while_count);
   stack_repeat = repeat_label = make_label ("WI$R", while_count);
   stack_recurse = recurse_label = make_label ("WI$S", while_count);

   count_index = get_symtab_temp(T_INT, T_SCALAR,T_NONE);
   sca_temp_start++; /* will prevent count_index temp from being used
		          when temps are reinitialized in body,loop and andbody */
   mreg_save = get_symtab_temp(T_LOGICAL,T_PARALLEL,W_LOGICAL);
   log_par_temp_start++; /* prevent mreg_save from being used */
   g_offset_start++;
	print_opcode ("begstack", T_BEGSTACK, T_NONE, T_NONE);
   print_args (symbol_table[count_index].tokpts,
	            symbol_table[mreg_save].tokpts, repeat_label, NULL, NULL);
   token_type = get_token();
   if (get_index(&invar_index,&invar_offset) == FAIL) return_value = FAIL;
   else if (token_type != T_IN)
     {print_msg (IN_ERR, token_buf);
      return_value = FAIL;}
   else
     {token_type = get_token();
      if(log_exp(&expr_index,&expr_offset) == FAIL)
         return_value = FAIL;
      else if (symbol_table[expr_index].mode2 != T_PARALLEL)
        {print_msg (PAREXP_ERR, NULL);
	      return_value = FAIL;}
      else if (symbol_table[expr_index].mode1 != T_LOGICAL)
        {print_msg (LOGEXP_ERR, NULL);
	      return_value = FAIL;}
      else
        {save_em(expr_index,expr_offset);
         print_opcode ("stackwhile", T_STACKWHILE, T_NONE, T_NONE);
      	print_args (symbol_table[count_index].tokpts,
							off_exp1(expr_index,expr_offset),
		               off_exp2(invar_index,invar_offset),
		               recurse_label, NULL);
         do_it(invar_index,invar_offset);
/*NOTICE- the unwind label in above is important difference from p1_while.c */
         if ((return_value = body()) != FAIL)
	        {if (token_type != T_ENDSTACKWHILE)
	            {print_msg (ENDSTACK_ERR, token_buf);
	             return_value = FAIL;}
	      else
	        {token_type = get_token();
	         if (strcmp(token_buf, symbol_table[invar_index].tokpts) != 0)
	           {print_msg (INXUNM_ERR, token_buf);
		         return_value = FAIL;}
	         else
	          {sca_temp_start--; /* restore temp starter */
              log_par_temp_start--; /* restore  mreg_save starter */
	           g_offset_start--;
			     print_opcode ("endstackwhile", T_ENDSTACKWHILE, T_NONE, T_NONE);
		        print_args (symbol_table[count_index].tokpts,
			                 symbol_table[mreg_save].tokpts,
			                 end_label,repeat_label, NULL, NULL, NULL);
             if ((token_type = get_token()) == T_LSQU)
		         {while ((token_type = get_token()) != T_RSQU);
		          token_type = get_token();}
		       if (token_type != T_EOS)
		         {print_msg (EOS_ERR, token_buf);
		          return_value = FAIL;}
		       else
		         {token_type = get_token();
		          return_value = SUCCESS;}
	      	}
	     }
	  }
   }
 }

   if (debug == 2)
      printf("While() returning return_value %d\n", return_value);

return(return_value);

} /* end of stackwhile_() */


