/* 910415  P1_SYMTAB.C */

#include "stdarg.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"
#include "alloc.h"
#include "p1_common.h"
#include "p1_parser.h"
#include "p1_scanner.h"
#include "errno.h"

#ifdef _WIN32
#include "dos.h"
#endif


/* ===========================================================================

   function name: symtab_insert
   input        : type          - token type
   mode1               - token mode (integer, real, etc.)
			    mode2               - token mode (scalar or parallel)
			    tok_size    - token size
			    string      - pointer to token buffer
			    dim_size    - array of dimension sizes
			    width               - width (in bits)
			    expt          - width of exponent for reals
		  signif        - width of mantisa for reals
			    num_dim     - number of dimensions
the following items are not part of the symtab_insert call, but the following
values are automatically inserted.
		  temp=FALSE    - TRUE if system temp
		  offset=0      - temp offset address
   output       : ---
   return value : ret_value     - FAIL, if error detected (duplicate entry,
						    out of memory, or symbol table overflow)
				      - symbol table index of token, otherwise

   global vars  : symbol_table  - symbol table data structure
			    debug               - debugging statement flag

   description  : This routine inserts an identifier into the symbol table.
			    The symbol table index where the token is to be inserted is
			    computed by the 'get_hash' function. If a collision occurs,
			    consecutive symbol table indices are examined until an empty
			    one is found. The function returns FAIL if :

		     1) the symbol table is full, or
		     2) no memory can be allocated for the token, or
		     3) the token has already been inserted.

		  The symbol table index of the token is returned otherwise.

=========================================================================== */

int symtab_insert(type,mode1,mode2,tok_size,string,dim_size,width,expt,signif,
		  num_dim)
int type, mode1, mode2, tok_size, width, num_dim, expt, signif;
char *string;
int *dim_size;
{
   int ret_value;             /* variable for the return value */
   int index;                 /* location of token in the symbol table */
   int symtab_count;    /* counter for number of locations in the symbol
				     table that have been passed */
   int errflag;         /* variable for error flag */
   int i;                        /* loop index */

  if (debug == 2)
      printf ("In insert().....token=%s, %x\n",string,string);

/* Initialize local variables. */
   index = get_hash(tok_size, string);
   symtab_count = 1;
   errflag = FALSE;

/*
 * While computed symbol table location is busy and no errors have occurred,
 * check if the token stored at the computed location matches the one to be
 * inserted. If so, signal an error and discontinue further searching.
 * Otherwise, a collision has occurred, so move to the next consecutive
 * symbol table location and repeat the process.
 */
   while (symbol_table[index].status == BUSY && !errflag)
     {if (strcmp(symbol_table[index].tokpts, string) == 0)
       if(WT == wt && module_flag == T_SUBROUTINE)
			{printf("WARNING - duplicate entry in subroutine\n");
			 ret_value = index;
			 errflag = TRUE;}
	    else
	{print_msg(DUP_ERR, string);
	 errflag = TRUE;
	 ret_value = FAIL;}
      else
	{index++;
	 if (index > NON_TEMP_ID_END)
		index = 0;
	 symtab_count++;
	 if (symtab_count > (NON_TEMP_ID_END + 1))
	   {print_msg(SYMTAB_ERR, string);
	    errflag = TRUE;
	    ret_value = FAIL;}
	}
      }

/*
 * If no errors were encountered, check if there is room to allocate
 * space for the new token and then fill the symbol table location.
 */
   if (!errflag)
     {if(tok_size <= 0) tok_size = 1; else tok_size += 1;
      symbol_table[index].tokpts =(char *)calloc(tok_size, sizeof(char));
      if (symbol_table[index].tokpts == NULL)
	{print_msg(MEM_ERR,NULL);
	 ret_value = FAIL;}
      else
	{strcpy(symbol_table[index].tokpts, string);
      symbol_table[index].status = BUSY;
	 symbol_table[index].mode1 = mode1;
	      symbol_table[index].mode2 = mode2;
	      symbol_table[index].toktype = type;
	      symbol_table[index].width = width;
	      symbol_table[index].expt = expt;
	      symbol_table[index].signif = signif;
	      symbol_table[index].num_dim = num_dim;
	      symbol_table[index].temp = FALSE;
	      symbol_table[index].offset = 0;
	 /* set to default use iospec_insert to overide */
	      symbol_table[index].iospec = shapedx;
			for (i = 0; i < num_dim; i++)
		 symbol_table[index].dim_size[i] = dim_size[i];
	      ret_value = index;
	}
     }
if(debug == 2) printf("Inserting %s, symbol_table[%d]=%s\n",string,index,
		       symbol_table[index].tokpts);
   if (debug == 2)
      printf ("Insert is returning %d\n", ret_value);
   return (ret_value);

} /* end of insert() */


/* ===========================================================================

   function name: get_hash
   input        : size          - token size
			    token               - token buffer pointer
   output       : ---
   return value : hash          - location/index of token in the symbol_table

   global vars   : debug                - debugging statement flag

   description  : This routine computes the index/location of a token in the
		  symbol table by using a hashing function based on the size
			    of the token and its spelling.

=========================================================================== */

int get_hash(size, token)
int size;           /* size of the token */
char *token;    /* token string */
{
   int hash;     /* hashed index of the token */

   hash = 0;
   switch (size)
     {case 1  : hash = hash + (int)token[0];
		break;
      case 2  : hash = hash + (int)token[size-1] * 32;
		hash = hash + (int)token[0];
		break;
      default : hash = (int)token[size/2] * 8;
		hash = hash + (int)token[size-1] * 32;
		hash = hash + (int)token[0];
      }
   hash = hash % (NON_TEMP_ID_END + 1);

   return (hash);

} /* end of get_hash() */


/* ===========================================================================

   function name: get_symtab_temp
   input               : result_mode1   - mode of result:
						int, real, logical, longint card
			    result_mode2        - mode of result: scalar or parallel
			    result_width        - width of result (in bits)
   output           : ---
   return_value : symtab_temp   - index of temporary variable, if any left
						  - FAIL, otherwise

   global vars  : symbol_table  - data structure for symbol table
			    sca_temp    - index of scalar temp var
			    ipar_temp   - index of int parallel temp var
			    rpar_temp   - index of real parallel temp var
			    lpar_temp   - index of logical parallel temp var
			    lipar_temp  - index of longint parallel temp var

   description  : This routine returns the index of a temporary variable
			    in the symbol table, if they have not all been used;
			    otherwise, FAIL is returned.

=========================================================================== */

int get_symtab_temp(result_mode1, result_mode2, result_size)
int result_mode1, result_mode2, result_size;
{
   int symtab_temp;             /* index of temporary variable */
	int inc;

  if (debug == 2)
      printf ("In get_symtab_temp(). result mode1=%d, mode2=%d, size=%d\n",
	      result_mode1, result_mode2, result_size);

   switch (result_mode2)
      {default: printf("%d ERROR - no temp storage for mode2 type %d.\n",
							  linenum,result_mode2);
		printf("            scalar assumed.\n");
      case T_CNST:     /* convert result_size to words for scalar */
      case T_REG_SCALAR:
      case T_SCALAR     : inc = (result_size-1)/SIZE_HOST+1;
		  if(result_mode1 == T_REAL)
		{if(real_sca_temp > REAL_SCA_TEMP_END)
		   {printf("%d real_sca_temp=%d, REAL_SCA_TEMP_START=%d, REAL_SCA_TEMP_END=%d\n",
		   linenum,real_sca_temp,REAL_SCA_TEMP_START,REAL_SCA_TEMP_END);
		    print_msg(TEMP_ERR, "in symta 1");
		    symtab_temp = FAIL;}
		 else
	     {symtab_temp = real_sca_temp;
	    if(result_size >= 1 )
		     real_sca_temp += inc;
	 else
	     real_sca_temp++;
				}
			 }
			    else /* result_mode1 == T_INT, among others */
			      if (sca_temp > INT_SCA_TEMP_END)
					  {print_msg(TEMP_ERR, "in symta 2");
					   symtab_temp = FAIL;}
					else
					  {symbol_table[sca_temp].mode1 = result_mode1;
					   symbol_table[sca_temp].mode2 = result_mode2;
					   symtab_temp = sca_temp;
			    if(result_size >= 1 )
					     sca_temp += inc;
									 else
					     sca_temp++;
									}
				      break;
      case T_PARALLEL: if (result_mode1 == T_INT ||
			   result_mode1 == T_LONGINT ||
					result_mode1 == T_OCT ||
					  result_mode1 == T_BIN ||
					  result_mode1 == T_HEX ||
					  result_mode1 == T_CHAR||
					  result_mode1 == T_CARD)
					{if (ipar_temp > INT_PAR_TEMP_END)
					   {print_msg(TEMP_ERR, "in symta 3");
						 symtab_temp = FAIL;}
					 else
					   {symbol_table[ipar_temp].mode1 = result_mode1;
						 symtab_temp = ipar_temp++;}
					}
		       else if (result_mode1 == T_REAL)
					{if (rpar_temp > REAL_PAR_TEMP_END)
					   {print_msg(TEMP_ERR, "in symta 4");
						 symtab_temp = FAIL;}
					 else
						symtab_temp = rpar_temp++;
					}
/* LITMP is used in em_* as divide and io routine temp area.
   use IPTMP for T_LONGINT
		       else if (result_mode1 == T_LONGINT)
					{if (lipar_temp > LINT_PAR_TEMP_END)
					   {print_msg(TEMP_ERR, "in symta 5);
						 symtab_temp = FAIL;}
					 else
						symtab_temp = lipar_temp++;
					}
*/
				      else if (result_mode1 == T_LOGICAL)
					{if (lpar_temp > LOG_PAR_TEMP_END)
					   {print_msg(TEMP_ERR, "in symta 6");
					    symtab_temp = FAIL;}
					 else
					   symtab_temp = lpar_temp++;
				   }
							 else printf("%d ERROR - unrecognized mode1 in get_symtab_temp\n",linenum);
			   break;
   }
/* offset is set in init for T_REAL */
if(result_mode1 != T_REAL)
  {symbol_table[symtab_temp].offset = g_offset;
   symbol_table[symtab_temp].temp = TRUE;
   symbol_table[symtab_temp].width = result_size; /* update size */
   if(result_mode2==T_PARALLEL) g_offset += result_size;}

return(symtab_temp);

} /* end of get_symtab_temp() */


/* ===========================================================================

   function name: symtab_fetch
   input               : size           - size of token
			    token               - token string
   output           : ---
   return value : ret_value     - index of token in symbol table, if found
						  - FAIL, otherwise

   global vars  : symbol_table  - data structure for symbol table

   description   : This routine finds the symbol table index of the input
			    token.

=========================================================================== */

int symtab_fetch(size, token)
int size;
char *token;
{
   int symtab_count;    /* number of symbol table locations searched */
   int ret_value;          /* return value */
   int index;              /* possible symbol table location of the token */
   int sindex;
   int done;               /* loop control variable */

// printf("In symtab_fetch(). size=%d, token=%s.\n",size,token);

   done = FALSE;
   ret_value = FAIL;
   index = get_hash(size, token);       /* compute hashed index of token */
   sindex = index;
   symtab_count = 1;

/* While the current symbol table index is not empty & no match has been
 * found for the token, continue checking successive contiguous locations.
 */
   while (!done && symbol_table[index].status == BUSY)
     {if (strcmp(symbol_table[index].tokpts, token) == 0)  /* if match...    */
	{done = TRUE;
	      ret_value = index;}                                  /* ..return index */
      else                                                 /* ..otherwise..  */
	{index++;                                          /* go to next loc */
	      if (index > (SYMTAB1_SIZE - 1))
	    index = 0;
	 symtab_count++;
	 if (symtab_count > SYMTAB1_SIZE)                   /* if no more...  */
		done = TRUE;                               /* ..exit loop    */
	}
     }

   /* this code loops around to top of table if symbol not found */
   if(sindex <= NON_TEMP_ID_END &&
		index  >= NON_TEMP_ID_END && !done)
     {index = 0;
      while(!done && symbol_table[index].status == BUSY)
	  {if(strcmp(symbol_table[index].tokpts,token) == 0)
			  {done = TRUE;
				ret_value = index;}
			else
			  {index++;
				if(index > sindex) done = TRUE;}
		  }
	  }

   if(strcmp("TRUE_",token) == 0)  {ret_value = LP_TRUE; token_type=T_ID;}
   if(strcmp("FALSE_",token) == 0) {ret_value = LP_FALSE; token_type=T_ID;}
   if(strcmp("STACK_",token) == 0) {ret_value = PARALLEL_STACK; token_type=T_ID;}

//printf("Symtab_fetch() returning index = %d\n", ret_value);

   return(ret_value);

} /* end of symtab_fetch() */


/* ===========================================================================

   function name: symtab_dump
   input               : ---
   output           : ---
   return value : ---

   global vars  : symbol_table  - array containing symbol table

   description   : This routine prints the contents of the non-empty
			    locations of the symbol table.

=========================================================================== */

void symtab_dump()
{
   int i;

   printf ("%-6s%-5s%-6s%-6s%-6s%-3s%-3s%-3s%-6s%-5s%-7s%-7s\n",
       "index","type","mode1","mode2","#dims","d1","d2","d3","width","expt",
	   "signif","string");
   printf ("==============================================================\n");
   for (i = 0; i < SYMTAB1_SIZE; i++)
     {if (symbol_table[i].status == BUSY)
	printf ("%-6d%-5d%-6d%-6d%-6d%-3d%-3d%-3d%-6d%-5d%-7d%s\n",
			   i,symbol_table[i].toktype,
			     symbol_table[i].mode1, symbol_table[i].mode2,
		   symbol_table[i].num_dim, symbol_table[i].dim_size[0],
			     symbol_table[i].dim_size[1], symbol_table[i].dim_size[2],
			     symbol_table[i].width, symbol_table[i].expt,
			     symbol_table[i].signif, symbol_table[i].tokpts);
      }
} /* end of symtab_dump() */


/* ===========================================================================

   function name: scalar_fetch
   input               : const_str              - string of the constant value
		  const_type              type of constant
   output           : ---
   return value : id_index              - symbol table index of the constant

   global vars  : token_type    - type of current token
			    token_buf   - current token string
			    symbol_table        - symbol table data structure
			    debug               - debugging statement flag

   description   : This routine fetch the symbol table index of the constant
			    by given a string of constant's value.
			    If the constant is not found, FAIL is returned
=========================================================================== */


int scalar_fetch(const_str,const_type)
char const_str[];
int const_type;
{
   int i;                       /* loop counter */
   int id_index;        /* symbol table index for scalar variable */
   int found;     /* found flag */
   char c_type[10];
   int mode1,mode2;
   char term;

   if (debug == 2)
      printf("In scalar_fetch(). Looking for %s......\n",const_str);

   i = 0;
   found = FALSE;

   switch(const_type)
     {case T_DEC_CONST:
      case T_OCT_CONST:
      case T_BIN_CONST:
      case T_HEX_CONST: mode1 = T_INT;
			mode2 = T_CNST;
			break;
		case T_STRING:    mode1 = T_CHAR;
			mode2 = T_CNST;
			break;
		case T_FLT_CONST: mode1 = T_REAL;
			mode2 = T_CNST;
			break;

      default: printf("%d ERROR - scalar_fetch, non integer constants %d not implemented.\n",
							  linenum, const_type);
     }
	  pc_itoa(mode1*SIZE_MODE2+mode2,c_type);
     if(wt == WT) term = '_'; else term = '$';
     while (i < total_scalars && found != TRUE)
     {/* are values equal && does the name end in $ */
      if (strcmp (scalar_table[i].arg3, const_str) == 0 &&
	       strcmp (scalar_table[i].arg4, c_type) == 0 &&
	 (scalar_table[i].arg1[strlen(scalar_table[i].arg1)-1] == term))
	if ((id_index = symtab_fetch(strlen(scalar_table[i].arg1),
							 scalar_table[i].arg1)) != FAIL)
	   if(symbol_table[id_index].toktype == const_type)
	      found = TRUE;
      i++;
  }
   if (debug == 2)
     printf("scalar_fetch() returning id_index : %d, found=%d\n", id_index,found);

   if(found == TRUE) return(id_index); else return(FAIL);

} /* end of scalar_fetch() */


/* ===========================================================================

   function name: get_const
   input               : const_str              - string of the constant value
		  const_type              type of the constant
   output        : ---
   return value : id_index              - symbol table index of the constant

   global vars  : token_type    - type of current token
			    token_buf   - current token string
			    symbol_table        - symbol table data structure
			    debug               - debugging statement flag

   description   : This routine fetches the symbol table index of the constant
			    given by const_str.
			    If the constant is not found, it is installed.
=========================================================================== */
long int atol();

int get_const(const_str,const_type)
char const_str[];
int const_type;
{
   char *const_name;
   int mode1,mode2;
   int size;                 /* loop counter */
   int id_index;        /* symbol table index for scalar variable */
   int shapedx;

  if (debug == 2)
    printf("In get_const(). Looking for %s, of type %d.\n",
					const_str,const_type);
   shapedx = FAIL;
   switch(const_type)
    {case T_STRING:    mode1 = T_INT;
		       mode2 = T_CNST;
		       if(strcmp(const_str,"") == 0) shapedx = SUCCESS;
    if( (id_index=scalar_fetch(const_str,const_type)) == FAIL)
       {const_name = make_constname(const_str,const_type);
	size = strlen(const_str)*8;
	if(symtab_fetch(strlen(const_name), const_name) == FAIL)
	     if((id_index = insert_scalar(const_type,T_CHAR, T_CNST,
		  strlen(const_name),const_name,size,0,size,const_str,shapedx)) == FAIL)
	   printf("%d ERROR - failure of insert scalar in p1_symtab\n", linenum);
		}
		       break;
	  case T_DEC_CONST:
     case T_OCT_CONST:
     case T_BIN_CONST:
     case T_HEX_CONST: mode1 = T_INT;
		       mode2 = T_CNST;
     if((id_index=scalar_fetch(const_str,const_type)) == FAIL)
       {const_name = make_constname(const_str,const_type);
       if(atol(const_str) <= 32767)
	  id_index = insert_scalar(const_type,mode1,mode2,
			  strlen(const_name),const_name,W_INT,0,W_INT,const_str,shapedx);
		  else
	  id_index = insert_scalar(const_type,mode1,mode2,
			  strlen(const_name),const_name,W_LONGINT,0,W_LONGINT,const_str,shapedx);
	if(module_flag == T_MAIN || wt_sub == WT)
	       {print_opcode("entry_",T_ENTRY,T_NONE,T_NONE);
		     print_args(const_name,NULL,NULL,NULL,NULL);}
	else if(code_type != EMULATE)
	       {print_opcode("extern_",T_EXTERN,T_NONE,T_NONE);
		     print_args(const_name,NULL,NULL,NULL,NULL);}
       }
		       break;
     case T_FLT_CONST: mode1 = T_REAL;
		       mode2 = T_CNST;
     if((id_index=scalar_fetch(const_str,const_type)) == FAIL)
       {const_name = make_constname(const_str,const_type);
	id_index = insert_scalar(const_type,mode1,mode2,
			  strlen(const_name),const_name,W_REAL,0,W_REAL,const_str,shapedx);
	if(module_flag == T_MAIN || wt_sub == WT)
	       {print_opcode("entry_",T_ENTRY,T_NONE,T_NONE);
		     print_args(const_name,NULL,NULL,NULL,NULL);}
	else if(code_type != EMULATE)
	       {print_opcode("extern_",T_EXTERN,T_NONE,T_NONE);
		     print_args(const_name,NULL,NULL,NULL,NULL);}
       }
		       break;

     default: printf("%d ERROR - unrecognized constant type %d.\n",
							 linenum, const_type);
		      break;
    }

  if (debug == 2)
      printf("get_const() returning id_index : %d\n", id_index);

   return(id_index);
}


/* ===========================================================================

   function name: define_fetch
   input               : define_str     - string of the define symbol
   output           : ---
   return value : def_index     - symbol table index of the define constant

   global vars  : token_type    - type of current token
			    token_buf   - current token string
			    symbol_table        - symbol table data structure
			    debug               - debugging statement flag

   description   : This routine fetchs the symbol table index for the given
			    string of define variable.

=========================================================================== */

int define_fetch(define_str)
char *define_str;
{
   int i;                       /* loop counter */
   int found;           /* flag if found */
   int def_index;       /* symbol table index for scalar variable
					 and be used for index in scalar table */

   if (debug == 2)
      printf("In define_fetch()....................................\n");

   i = 0;
   found = FALSE;

   while (i < total_scalars && found != TRUE)
     {if (strcmp (scalar_table[i].arg1, define_str) == 0)
	 if ((def_index = symtab_fetch(strlen(scalar_table[i].arg1),
													scalar_table[i].arg1)) != FAIL)
	    found = TRUE;
      i++;
     }

   if(found)
     def_index = i -1;
   else
     def_index = FAIL;

   if(debug == 2)
      printf("define_fetch() returning id_index : %d\n", def_index);

   return(def_index);

} /* end of define_fetch() */


/* insert_scalar used only by define constant & therefore are never temps.
   That is, insert_scalar calls insert scalar with FALSE always  */
insert_scalar(type,mode1,mode2,name_length,name,field_length,expt_length,
	      signif_length,value,shape)
int type,mode1,mode2,name_length,field_length,expt_length,signif_length;
char *name,*value;
int shape;

{ int id_index;
  char field[10], expt[10], signif[10], c_mode[10];
  int foo[10];

  id_index = symtab_insert(type,mode1,mode2,name_length,name,foo,field_length,
			   expt_length,signif_length,0);
  pc_itoa(mode1*SIZE_MODE2+mode2,c_mode);
  pc_itoa(field_length,field);
  if(mode1 == T_REAL)
    {pc_itoa(expt_length,expt); pc_itoa(signif_length,signif);}
  else
    {pc_itoa(0,expt); pc_itoa(signif_length+expt_length,signif);}
  install_scalar(name,"1",value,c_mode,field,expt,signif,FALSE,shape);

  return(id_index);
}


char  *index(); /* declare index to be pointer type */

/* off_exp generates the string consisting of the identifier
   and its offset, i.e.  id,offset.  There are three of them
   one for each argument of a quadruple.
   Also converts ?TMP$+n to ?TMP$+offset if symbol_table.temp is TRUE    */


char *off_exp1(indx,offset)
int indx,offset;
{
  char *pnt,c_offset[10];

  strcpy(buffer1,symbol_table[indx].tokpts);
  if((pnt=index(buffer1,'+')) != NULL    &&
     symbol_table[indx].temp             &&
	  (symbol_table[indx].mode2 == T_PARALLEL))
    {*(pnt+1) = 0;
     pc_itoa(symbol_table[indx].offset,c_offset);
     strcat(buffer1,c_offset); }

  if(offset != SUCCESS)
    {strcat(buffer1,",");
     strcat(buffer1,symbol_table[offset].tokpts);}
  return (buffer1);
}

char *off_exp2(indx,offset)
int indx,offset;
{
  char *pnt,c_offset[10];

  strcpy(buffer2,symbol_table[indx].tokpts);

  if(symbol_table[indx].temp  && (symbol_table[indx].mode2 == T_PARALLEL) &&
     (pnt=index(buffer2,'+')) != NULL)
    {*(pnt+1) = 0;
     pc_itoa(symbol_table[indx].offset,c_offset);
     strcat(buffer2,c_offset); }

  if(offset != SUCCESS)
    {strcat(buffer2,",");
     strcat(buffer2,symbol_table[offset].tokpts);}
  return (buffer2);
}

char *off_exp3(indx,offset)
int indx,offset;
{
  char *pnt,c_offset[10];

  strcpy(buffer3,symbol_table[indx].tokpts);

  if(symbol_table[indx].temp  && (symbol_table[indx].mode2 == T_PARALLEL) &&
     (pnt = index(buffer3,'+')) != NULL)
    {*(pnt+1) = 0;
     pc_itoa(symbol_table[indx].offset,c_offset);
     strcat(buffer3,c_offset); }

  if(offset != SUCCESS)
    {strcat(buffer3,",");
     strcat(buffer3,symbol_table[offset].tokpts);}
  return (buffer3);
}

char *off_exp4(indx,offset)
int indx,offset;
{
  char *pnt,c_offset[10];

  strcpy(buffer4,symbol_table[indx].tokpts);

  if(symbol_table[indx].temp  && (symbol_table[indx].mode2 == T_PARALLEL) &&
     (pnt = index(buffer4,'+')) != NULL)
    {*(pnt+1) = 0;
     pc_itoa(symbol_table[indx].offset,c_offset);
     strcat(buffer4,c_offset); }

  if(offset != SUCCESS)
    {strcat(buffer4,",");
     strcat(buffer4,symbol_table[offset].tokpts);}
  return (buffer4);
}

/* biggest_size selects the biggest size from index1 and index2 */
/* if both are parallel or scalar, if only one is parallel, it picks it */
biggest_size(index1,index2)
int index1,index2;
{
  int size1,size2,size3;

  size1 = symbol_table[index1].width;
  size2 = symbol_table[index2].width;

  if(symbol_table[index1].mode2==symbol_table[index2].mode2 ||
     symbol_table[index1].mode2==T_CNST                     ||
     symbol_table[index2].mode2==T_CNST )
    if (size1 < size2) size3=size2; else size3=size1;
  else if (symbol_table[index1].mode2 == T_PARALLEL)
    size3=size1;
  else size3=size2;

return(size3);
}

iospec_insert(index,specdx)
int index;
int specdx;
{
   if(index != FAIL)
		symbol_table[index].iospec =  specdx;
	else
		printf("%d ERROR - attempt to add iospec to illegal index\n",linenum);
return 0;

}

/* to32 modify to free */
/*
void far_symtab_free()
{
   int i;
printf("Before free, farcore=%d\n",farcoreleft());
   for (i = 0; i < SYMTAB1_SIZE; i++)
     if (symbol_table[i].status == BUSY)
       if(*symbol_table[i].tokpts != 0)
	   {farfree(symbol_table[i].tokpts);
	    symbol_table[i].status = IDLE;}
printf("After free, farcore=%d\n",farcoreleft());
   if(farheapcheck() <0)
     printf("ERROR - farheap corrupted in far_symtab_free\n");
}
*/
